"""
alertmanagermeshtastic.meshtastic
~~~~~~~~~~~~~~~

Meshtastic connection

:Copyright: 2007-2022 Jochen Kupperschmidt
:Copyright: 2023 Alexander Volz
:License: MIT, see LICENSE for details.
"""

from __future__ import annotations
import logging
import meshtastic, meshtastic.serial_interface

from dateutil import parser
from .config import MeshtasticConfig, MeshtasticConnection
import time

from pubsub import pub

logger = logging.getLogger(__name__)


class Announcer:
    """An announcer."""

    def start(self) -> None:
        """Start the announcer."""

    def announce(self, alert: str) -> None:
        """Announce a message."""
        raise NotImplementedError()

    def shutdown(self) -> None:
        """Shut the announcer down."""


class MeshtasticAnnouncer(Announcer):
    """An announcer that writes messages to MESHTASTIC."""

    def __init__(
        self,
        connection: MeshtasticConnection,
    ) -> None:
        self.connection = connection

        self.meshtasticinterface = _create_meshtasticinterface(connection)

    def _onconnect(topic=pub.AUTO_TOPIC, interface=None):
        logger.debug("\t Connected to meshtastic")

    def _onconnectionlost(topic=pub.AUTO_TOPIC, interface=None):
        del topic.meshtasticinterface
        while True:
            try:
                topic.meshtasticinterface = _create_meshtasticinterface(
                    topic.connection
                )
                break
            except Exception as e:
                logger.error(
                    "\t Connnection to meshtastic failed with error: %s , retry in 2 seconds",
                    e,
                )
                time.sleep(2)

    def start(self) -> None:
        """Connect to the connection, in a separate thread."""
        logger.info(
            '\t Connecting to MESHTASTIC connection %s, the node is %d and messages will be sent %d times with timeout %d before failing',
            self.connection.tty,
            self.connection.nodeid,
            self.connection.maxsendingattempts,
            self.connection.timeout,
        )
        pub.subscribe(self._onconnectionlost, "meshtastic.connection.lost")
        pub.subscribe(self._onconnect, "meshtastic.connection.established")
        # start_thread(self.meshtasticinterface.start)

    def announce(self, alert: str) -> None:
        """Announce a message."""
        try:
            try:
                message = self.formatalert(alert)

            except Exception as e:
                logger.error(
                    "\t [%s][%d] Message formatting failed: %s",
                    alert["fingerprint"],
                    alert["qn"],
                    e,
                )
                raise

            try:
                chunks = self.splitmessagesifnessecary(message, alert)
                total_chunks = len(chunks)
                logger.debug(
                    "\t [%s][%d] splitted in %d chunks",
                    alert["fingerprint"],
                    alert["qn"],
                    total_chunks,
                )

            except Exception as e:
                logger.error(
                    "\t [%s][%d] could not split in chunks: %s",
                    alert["fingerprint"],
                    alert["qn"],
                    e,
                )
                raise

            for index, chunk in enumerate(chunks):
                for attempt in range(self.connection.maxsendingattempts):
                    logger.debug(
                        "\t [%s][%d][%d] sending attempt %d ",
                        alert["fingerprint"],
                        alert["qn"],
                        index,
                        attempt,
                    )
                    try:
                        while not hasattr(self, 'meshtasticinterface'):
                            time.sleep(2)

                        self.meshtasticinterface.sendText(
                            str(alert["qn"])
                            + ":"
                            + str(index + 1)
                            + "/"
                            + str(total_chunks)
                            + "\n"
                            + chunk,
                            self.connection.nodeid,
                            True,
                            False,
                            self.meshtasticinterface.getNode(
                                self.connection.nodeid, False
                            ).onAckNak,
                        )
                        self.meshtasticinterface.waitForAckNak()
                        start_time = time.time()
                        while True:
                            # Check if value is True
                            if (
                                self.meshtasticinterface._acknowledgment.receivedAck
                            ):
                                logger.debug(
                                    "\t [%s][%d][%d] got ack received from meshtastic on attempt %d",
                                    alert["fingerprint"],
                                    alert["qn"],
                                    index,
                                    attempt,
                                )
                                break

                            # Check if timeout has been reached
                            if (
                                time.time() - start_time
                                > self.connection.timeout
                            ):
                                logger.debug(
                                    "\t [%s][%d][%d] Timeout reached on attempt %d!",
                                    alert["fingerprint"],
                                    alert["qn"],
                                    index,
                                    attempt,
                                )
                                raise Exception(
                                    "No ack received from meshtastic within the timeout"
                                )
                                break

                            # Sleep for a short period to avoid a busy wait
                            time.sleep(0.1)

                        logger.debug(
                            "\t [%s][%d][%d] success on attempt %d  ",
                            alert["fingerprint"],
                            alert["qn"],
                            index,
                            attempt,
                        )
                        break
                    except Exception as e:
                        logger.error(
                            "\t [%s][%d][%d] failed on attempt %d with error: %s",
                            alert["fingerprint"],
                            alert["qn"],
                            index,
                            attempt,
                            e,
                        )
                        if attempt == self.connection.maxsendingattempts - 1:
                            raise

        except Exception as e:
            logger.error(
                "\t [%s][%d] send Attempt failed with error: %s",
                alert["fingerprint"],
                alert["qn"],
                e,
            )

    def splitmessagesifnessecary(self, message, alert):
        chunk_size = 160
        if len(message) > chunk_size:
            logger.debug(
                "\t [%s][%d] Message to big, split to chunks",
                alert["fingerprint"],
                alert["qn"],
            )
            chunks = [
                message[i : i + chunk_size]
                for i in range(0, len(message), chunk_size)
            ]
            return chunks
        else:
            logger.debug(
                "\t [%s][%d] Message size okay",
                alert["fingerprint"],
                alert["qn"],
            )
            return [message]

    def formatalert(self, alert):
        message = "Status: " + alert["status"] + "\n"
        if "name" in alert["labels"]:
            message += (
                "Instance: "
                + alert["labels"]["instance"]
                + "("
                + alert["labels"]["name"]
                + ")\n"
            )
        elif "instance" in alert["labels"]:
            message += "Instance: " + alert["labels"]["instance"] + "\n"
        elif "alertname" in alert["labels"]:
            message += "Alert: " + alert["labels"]["alertname"] + "\n"
        if "info" in alert["annotations"]:
            message += "Info: " + alert["annotations"]["info"] + "\n"
        if "summary" in alert["annotations"]:
            message += "Summary: " + alert["annotations"]["summary"] + "\n"
        # if 'description' in alert['annotations']:
        #     message += (
        #         "Description: " + alert['annotations']['description'] + "\n"
        #     )
        if alert["status"] == "resolved":
            correctdate = parser.parse(alert["endsAt"]).strftime(
                "%Y-%m-%d %H:%M:%S"
            )
            message += "Resolved: " + correctdate
        elif alert["status"] == "firing":
            correctdate = parser.parse(alert["startsAt"]).strftime(
                "%Y-%m-%d %H:%M:%S"
            )
            message += "Started: " + correctdate
        return message

    def shutdown(self) -> None:
        """Shut the announcer down."""
        self.meshtasticinterface.close()


class Meshtasticinterface(meshtastic.serial_interface.SerialInterface):
    """An MESHTASTIC Interface to forward messages to MESHTASTIC devices."""

    def get_version(self) -> str:
        """Return this on CTCP VERSION requests."""
        return 'alertmanagermeshtastic'


def _create_meshtasticinterface(
    connection: MeshtasticConnection,
) -> Meshtasticinterface:
    """Create a Interface."""

    meshtasticinterface = Meshtasticinterface(connection.tty)

    return meshtasticinterface


class DummyAnnouncer(Announcer):
    """An announcer that writes messages to STDOUT."""

    def announce(self, alert: str) -> None:
        """Announce a message."""
        logger.debug('%s> %s', alert)


def create_announcer(config: MeshtasticConfig) -> Announcer:
    """Create an announcer."""
    if config.connection is None:
        logger.info(
            '\t No MESHTASTIC connection specified; will write to STDOUT instead.'
        )
        return DummyAnnouncer()

    return MeshtasticAnnouncer(
        config.connection,
    )



"""
:Copyright: 2007-2022 Jochen Kupperschmidt
:License: MIT, see LICENSE for details.
"""

import pytest

from alertmanagermeshtastic.meshtastic import MeshtasticChannel


@pytest.mark.parametrize(
    'channel, expected_name, expected_password',
    [
        (MeshtasticChannel('#example'),                         '#example',      None    ),
        (MeshtasticChannel('#example', password=None),          '#example',      None    ),
        (MeshtasticChannel('#headquarters', password='secret'), '#headquarters', 'secret'),
    ],
)
def test_meshtastic_channel_creation(channel, expected_name, expected_password):
    assert channel.name == expected_name
    assert channel.password == expected_password



"""
:Copyright: 2007-2022 Jochen Kupperschmidt
:License: MIT, see LICENSE for details.
"""

import pytest

from alertmanagermeshtastic.meshtastic import (
    create_announcer,
    DummyAnnouncer,
    MeshtasticAnnouncer,
    MeshtasticConfig,
    MeshtasticServer,
)


@pytest.mark.parametrize(
    'server, expected_type',
    [
        (MeshtasticServer('meshtastic.server.test'), MeshtasticAnnouncer),
        (None, DummyAnnouncer),
    ],
)
def test_create_announcer(server, expected_type):
    config = MeshtasticConfig(
        server=server,
        nickname='nick',
        realname='Nick',
        commands=[],
        channels=set(),
    )

    announcer = create_announcer(config)

    assert type(announcer) == expected_type



"""
alertmanagermeshtastic.signals
~~~~~~~~~~~~~~~~~~~

Signals

:Copyright: 2007-2022 Jochen Kupperschmidt
:Copyright: 2023 Alexander Volz
:License: MIT, see LICENSE for details.
"""

from blinker import Signal


message_received = Signal()



"""
alertmanagermeshtastic.config
~~~~~~~~~~~~~~~~~~

Configuration loading

:Copyright: 2007-2022 Jochen Kupperschmidt
:Copyright: 2023 Alexander Volz
:License: MIT, see LICENSE for details.
"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional

import rtoml

DEFAULT_HTTP_HOST = '127.0.0.1'
DEFAULT_HTTP_PORT = 9119
DEFAULT_MESHTASTIC_NODEID = 123456789
DEFAULT_MESHTASTIC_MAXSENDINGATTEMPTS = 5
DEFAULT_MESHTASTIC_TIMEOUT = 60


class ConfigurationError(Exception):
    """Indicates a configuration error."""


@dataclass(frozen=True)
class Config:
    log_level: str
    http: HttpConfig
    meshtastic: MeshtasticConfig


@dataclass(frozen=True)
class HttpConfig:
    """An HTTP receiver configuration."""

    host: str
    port: int


@dataclass(frozen=True)
class MeshtasticConnection:
    """An MESHTASTIC connection."""

    tty: str
    nodeid: int = DEFAULT_MESHTASTIC_NODEID
    maxsendingattempts: int = DEFAULT_MESHTASTIC_MAXSENDINGATTEMPTS
    timeout: int = DEFAULT_MESHTASTIC_TIMEOUT


@dataclass(frozen=True)
class MeshtasticConfig:
    """An MESHTASTIC Interface configuration."""

    connection: Optional[MeshtasticConnection]


def load_config(path: Path) -> Config:
    """Load configuration from file."""
    data = rtoml.load(path)

    log_level = _get_log_level(data)
    http_config = _get_http_config(data)
    meshtastic_config = _get_meshtastic_config(data)

    return Config(
        log_level=log_level,
        http=http_config,
        meshtastic=meshtastic_config,
    )


def _get_log_level(data: dict[str, Any]) -> str:
    level = data.get('log_level', 'debug').upper()

    if level not in {'CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG'}:
        raise ConfigurationError(f'Unknown log level "{level}"')

    return level


def _get_http_config(data: dict[str, Any]) -> HttpConfig:
    data_http = data.get('http', {})

    host = data_http.get('host', DEFAULT_HTTP_HOST)
    port = int(data_http.get('port', DEFAULT_HTTP_PORT))

    return HttpConfig(host, port)


def _get_meshtastic_config(data: dict[str, Any]) -> MeshtasticConfig:
    data_meshtastic = data['meshtastic']

    connection = _get_meshtastic_connection(data_meshtastic)

    return MeshtasticConfig(
        connection=connection,
    )


def _get_meshtastic_connection(
    data_meshtastic: Any,
) -> Optional[MeshtasticConnection]:
    data_connection = data_meshtastic.get('connection')
    if data_connection is None:
        return None

    maxsendingattempts = data_connection.get('maxsendingattempts')
    timeout = data_connection.get('timeout')
    tty = data_connection.get('tty')
    if not tty:
        return None

    nodeid = int(data_connection.get('nodeid', DEFAULT_MESHTASTIC_NODEID))

    return MeshtasticConnection(
        tty=tty,
        nodeid=nodeid,
        maxsendingattempts=maxsendingattempts,
        timeout=timeout,
    )



"""
:Copyright: 2007-2022 Jochen Kupperschmidt
:License: MIT, see LICENSE for details.
"""

from pathlib import Path

import pytest

from alertmanagermeshtastic.cli import parse_args


def test_parse_args_without_args():
    with pytest.raises(SystemExit):
        actual = parse_args([])


def test_parse_args_with_config_filename_arg():
    actual = parse_args(['config.toml'])

    assert actual.config_filename == Path('config.toml')



"""
alertmanagermeshtastic.util
~~~~~~~~~~~~~~~~

Utilities

:Copyright: 2007-2022 Jochen Kupperschmidt
:License: MIT, see LICENSE for details.
"""

import logging
from logging import Formatter, StreamHandler
from threading import Thread
from typing import Callable, Optional


def configure_logging(level: str) -> None:
    """Configure application-specific loggers.

    Setting the log level does not affect dependencies' loggers.
    """
    # Get the parent logger of all application-specific
    # loggers defined in the package's modules.
    pkg_logger = logging.getLogger(__package__)

    # Configure handler that writes to STDERR.
    handler = StreamHandler()
    handler.setFormatter(Formatter('%(asctime)s %(levelname)-8s %(message)s'))
    pkg_logger.addHandler(handler)

    pkg_logger.setLevel(level)


def start_thread(target: Callable, name: Optional[str] = None) -> None:
    """Create, configure, and start a new thread."""
    t = Thread(target=target, name=name, daemon=True)
    t.start()



"""
alertmanagermeshtastic.http
~~~~~~~~~~~~~~~~

HTTP server to receive messages

:Copyright: 2007-2022 Jochen Kupperschmidt
:Copyright: 2023 Alexander Volz
:License: MIT, see LICENSE for details.
"""

from __future__ import annotations
from http import HTTPStatus
import logging
import sys
from wsgiref.simple_server import make_server, ServerHandler, WSGIServer

from werkzeug.exceptions import abort, HTTPException
from werkzeug.routing import Map, Rule
from werkzeug.wrappers import Request, Response

from .config import HttpConfig
from .signals import message_received
from .util import start_thread


logger = logging.getLogger(__name__)


def create_app() -> Application:
    return Application()


class Application:
    def __init__(
        self,
    ) -> None:
        self._url_map = Map(
            [
                Rule('/alert', endpoint='alert', methods=['POST']),
            ]
        )

    def __call__(self, environ, start_response):
        return self.wsgi_app(environ, start_response)

    def wsgi_app(self, environ, start_response):
        request = Request(environ)
        response = self.dispatch_request(request)
        return response(environ, start_response)

    def dispatch_request(self, request: Request):
        adapter = self._url_map.bind_to_environ(request.environ)

        try:
            endpoint, values = adapter.match()
            handler = getattr(self, f'on_{endpoint}')
            return handler(request, **values)
        except HTTPException as exc:
            return exc

    def on_alert(self, request: Request) -> Response:
        try:
            data = _extract_payload(request, {'alerts'})

            for alert in data["alerts"]:
                logger.debug("\t [%s] received", alert["fingerprint"])
                message_received.send(alert=alert)
            return Response('Alert OK', status=HTTPStatus.OK)
        except Exception as error:
            logger.error("\t could not queue alerts: %s ", error)
            return Response('Alert fail', status=HTTPStatus.OK)


def _extract_payload(request: Request, keys: set[str]) -> dict[str, str]:
    """Extract values for given keys from JSON payload."""
    if not request.is_json:
        abort(HTTPStatus.UNSUPPORTED_MEDIA_TYPE)

    payload = request.json
    if payload is None:
        abort(HTTPStatus.BAD_REQUEST)

    data = {}
    try:
        for key in keys:
            data[key] = payload[key]
    except KeyError:
        abort(HTTPStatus.BAD_REQUEST)

    return data


# Override value of `Server:` header sent by wsgiref.
ServerHandler.server_software = 'alertmanagermeshtastic'


def create_server(config: HttpConfig) -> WSGIServer:
    """Create the HTTP server."""
    app = create_app()

    return make_server(config.host, config.port, app)


def start_receive_server(config: HttpConfig) -> None:
    """Start in a separate thread."""
    try:
        server = create_server(config)
    except OSError as e:
        sys.stderr.write(f'Error {e.errno:d}: {e.strerror}\n')
        sys.stderr.write(
            f'Probably no permission to open port {config.port}. '
            'Try to specify a port number above 1,024 (or even '
            '4,096) and up to 65,535.\n'
        )
        sys.exit(1)

    thread_name = server.__class__.__name__
    start_thread(server.serve_forever, thread_name)
    logger.info('Listening for HTTP requests on %s:%d.', *server.server_address)



"""
alertmanagermeshtastic.processor
~~~~~~~~~~~~~~~~~~~~~

Connect HTTP server and MESHTASTIC interface.

:Copyright: 2007-2022 Jochen Kupperschmidt
:Copyright: 2023 Alexander Volz
:License: MIT, see LICENSE for details.
"""

from __future__ import annotations
import logging
from queue import SimpleQueue
from typing import Any, Optional

from .config import Config
from .http import start_receive_server
from .meshtastic import create_announcer
from .signals import message_received


logger = logging.getLogger(__name__)


class Processor:
    def __init__(self, config: Config) -> None:
        self.config = config
        self.announcer = create_announcer(config.meshtastic)
        self.enabled_channel_names: set[str] = set()
        self.message_queue: SimpleQueue = SimpleQueue()
        self.qn = 0

        # Up to this point, no signals must have been sent.
        self.connect_to_signals()
        # Signals are allowed be sent from here on.

    def connect_to_signals(self) -> None:
        message_received.connect(self.handle_message)

    def handle_message(
        self,
        sender: Optional[Any],
        *,
        alert: str,
    ) -> None:
        """Log and announce an incoming message."""
        self.qn = self.qn + 1
        alert["qn"] = self.qn
        logger.debug(
            '\t [%s][%d] put in queue',
            alert["fingerprint"],
            alert["qn"],
        )
        self.message_queue.put((alert))

    def announce_message(self, alert: str) -> None:
        """Announce message on MESHTASTIC."""
        self.announcer.announce(alert)

    def process_queue(self, timeout_seconds: Optional[int] = None) -> None:
        logger.debug('\t Messages in queue: %d', self.message_queue.qsize())
        """Process a message from the queue."""
        alert = self.message_queue.get(timeout=timeout_seconds)
        logger.debug(
            '\t [%s][%d] processing message ', alert["fingerprint"], alert["qn"]
        )
        self.announce_message(alert)

    def run(self) -> None:
        """Run the main loop."""
        self.announcer.start()
        start_receive_server(self.config.http)

        logger.info('\t Starting to process queue ...')
        try:
            while True:
                self.process_queue()
        except KeyboardInterrupt:
            pass

        logger.info('\t Shutting down ...')
        self.announcer.shutdown()


def start(config: Config) -> None:
    """Start the MESHTASTIC interface and the HTTP listen server."""
    processor = Processor(config)
    processor.run()



"""
alertmanagermeshtastic.cli
~~~~~~~~~~~~~~~

Command line entry point

:Copyright: 2007-2022 Jochen Kupperschmidt
:Copyright: 2023 Alexander Volz
:License: MIT, see LICENSE for details.
"""

from __future__ import annotations
from argparse import ArgumentParser, Namespace
from pathlib import Path
import sys

from . import VERSION
from .config import load_config
from .processor import start
from .util import configure_logging


def parse_args(args: list[str]) -> Namespace:
    """Parse command line arguments."""
    parser = ArgumentParser()
    parser.add_argument(
        '--version',
        action='version',
        version=f'alertmanagermeshtastic {VERSION}',
    )
    parser.add_argument('config_filename', type=Path)
    return parser.parse_args(args)


def main() -> None:
    """Load the configuration file, start the MESHTASTIC Interface and HTTP listen server."""
    namespace = parse_args(sys.argv[1:])
    config = load_config(namespace.config_filename)
    configure_logging(config.log_level)
    start(config)


if __name__ == '__main__':
    main()
